#include <sys/termios.h> // tcgetattr
#include <stdio.h>
#include <sys/select.h>
#include <alloca.h>
#include <stdarg.h>
#include <getopt.h>
#include <time.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <fcntl.h>
#include "ts732.h"
#include "file.h"
#include "serial.h"

// copyright (c)2006 Technologic Systems
// Author: Michael Schmidt

int TS732requestInitFromReply(ByteArrayRef bRep,ByteArrayRef bReq);

void AVRadcVoltsPrint(int n,word val,word min,word max,word tpv,word off) {
  printf("ADC%d      = %.2fV (%.2fV..%.2fV) [%04X,%04X..%04X]\n",n,
	 AVRcalcADCvolts(val,tpv,off),
	 AVRcalcADCvolts(min,tpv,off),
	 AVRcalcADCvolts(max,tpv,off),
	 val,min,max
	 );
}

inline word AVRcalcTimeFactor(dword ARMseconds,dword AVRseconds) {
  return (word)(ARMseconds * 32768 / AVRseconds);
}

inline dword AVRcalcARMseconds(dword AVRseconds, word timeFactor) {
  return timeFactor * AVRseconds / 32768;
}

inline dword AVRcalcAVRseconds(dword ARMseconds, word timeFactor) {
  return ARMseconds * 32768 / timeFactor;
}

/*
inline word AVRcalcDACmax(word highTimeOut,word measV) {
  return highTimeOut * 10 / measV;
}
*/

inline dword AVRcalcDAChighTime(word DACmax,word targetV) {
  return (word)((dword)DACmax * targetV / 65536);
}

inline dword AVRcalcTPV(word adcV2,word adcV1,word V2,word V1) {
  return 2 * (adcV2 - adcV1) / (V2-V1);
}

inline dword AVRcalcOff(word adcV2, word adcV1,word V2,word V1) {
  return adcV1 - (AVRcalcTPV(adcV2,adcV1,V2,V1) * V1 / 2);
}

inline float AVRcalcADCvolts(word adcV,word adcTPV,word adcOff) {
  return (float)(adcV - adcOff) / adcTPV;
}

void AVRrequestPrint(AVRrequest *r) {
  printf("alarm=%d\n",r->alarm);
  printf("DAC1en=%d\n",r->DAC1en);
  printf("DAC2en=%d\n",r->DAC2en);
  printf("_LEDbot=%d\n",r->_LEDbot);
  printf("_LEDtop=%d\n",r->_LEDtop);
  printf("EEwrite=%d\n",r->EEwrite);
  printf("EEread=%d\n",r->EEread);
  printf("DDR5=%d\n",r->DDR5);
  printf("DDR6=%d\n",r->DDR6);
  printf("DDR7=%d\n",r->DDR7);
  printf("DDR8=%d\n",r->DDR8);
  printf("DR5=%d\n",r->DR5);
  printf("DR6=%d\n",r->DR6);
  printf("DR7=%d\n",r->DR7);
  printf("DR8=%d\n",r->DR8);
  printf("EEdata=%d\n",r->EEdata);
  printf("EEadrs=%d\n",r->EEadrs);
  printf("ADC2filter=%d\n",r->ADC2filter);
  printf("ADC3filter=%d\n",r->ADC3filter);
  printf("ADC4filter=%d\n",r->ADC4filter);
  printf("ADC6filter=%d\n",r->ADC6filter);
  printf("ADC1filter=%d\n",r->ADC1filter);
}

void AVRreplyPrint(AVRreply *r) {
  printf("millitime=%d\n",r->millitime);
  printf("DAC1en=%d\n",r->DAC1en);
  printf("DAC2en=%d\n",r->DAC2en);
  printf("_LEDbot=%d\n",r->_LEDbot);
  printf("_LEDtop=%d\n",r->_LEDtop);
  printf("EEwrite=%d\n",r->EEwrite);
  printf("EEread=%d\n",r->EEread);
  printf("DDR5=%d\n",r->DDR5);
  printf("DDR6=%d\n",r->DDR6);
  printf("DDR7=%d\n",r->DDR7);
  printf("DDR8=%d\n",r->DDR8);
  printf("DR5=%d\n",r->DR5);
  printf("DR6=%d\n",r->DR6);
  printf("DR7=%d\n",r->DR7);
  printf("DR8=%d\n",r->DR8);
  printf("EEdata=%d\n",r->EEdata);
  printf("EEadrs=%d\n",r->EEadrs);
  printf("ADC2filter=%d\n",r->ADC2filter);
  printf("ADC3filter=%d\n",r->ADC3filter);
  printf("ADC4filter=%d\n",r->ADC4filter);
  printf("ADC6filter=%d\n",r->ADC6filter);
  printf("ADC1filter=%d\n",r->ADC1filter);
}

/*
  function      resend             return
  readEEPROM    EEadrs != adrs     EEready
  writeEEPROM   (*)                (*)
  DAC1en        DAC1en != en       1
  DAC2en        _ != __
  _LEDbot         "
  _LEDtop         "
  DDR(n)          "
  DR(n)           " ?
  ADC(n)filter    "
  alarm         if ANY reply received

  (*) we must perform a readEEPROM call.  if the value read back matches
  what we wrote, then we are done.
*/

int AVREEPROMdoneCallback(ByteArrayRef req1,ByteArrayRef rep1) {
  AVRrequest *req = (AVRrequest *)req1.arr;
  AVRreply *rep = (AVRreply *)rep1.arr;

  if (rep->EEready == 0) {
    return 0;
  }
  return 1;
}

/*
  To reduce the chance of needlessly writing to the EEPROM, we should
  pass an opaque data structure to the callback, which the callback
  could use to maintain a counter.  It would then only indicate it
  wanted a retry after the counter expired.
  For now, we wanted to simply throw away the first reply from the AVR, 
  however this seems to slow things down, sometimes significantly, perhaps
  because sometimes the first retry isn't needed.
 */
int AVREEPROMcallback(ByteArrayRef req1,ByteArrayRef rep1) {
  AVRrequest *req = (AVRrequest *)req1.arr;
  AVRreply *rep = (AVRreply *)rep1.arr;

  if (rep->EEadrs != req->EEadrs) {
    return 1;
  }
  if (rep->EEready) {
    return 0;
  }
  return 2;
}

#define BYTE_ARRAY_REF_STRUCT(name) BAR((byte *)name,sizeof(*(name)))

void AVRwriteEEPROM(FILE *f,byte adrs,byte value) {
  AVRreply pkt;
  AVRrequest req;
  ByteArrayRef b = BYTE_ARRAY_REF_STRUCT(&pkt);

  fileFlushInput(f);
  b = AVRreplyGet(f,b);
  TS732requestInitFromReply(BYTE_ARRAY_REF_STRUCT(&pkt),
			    BYTE_ARRAY_REF_STRUCT(&req));
  req.EEadrs = adrs;
  req.EEdata = value;
  req.EEwrite = 1;
  req.EEread = 0;

  printf("EE WR %02X -> [%02X]\n",value,adrs);
  AVRsendRequest(f,BYTE_ARRAY_REF_STRUCT(&req),BYTE_ARRAY_REF_STRUCT(&pkt),
		 &AVREEPROMcallback);
}

byte AVRreadEEPROM(FILE *f,byte adrs,AVRreply *pkt) {
  AVRreply pkt2;
  AVRrequest req;
  ByteArrayRef b = BYTE_ARRAY_REF_STRUCT(pkt);

  b = AVRreplyGet(f,b);
  TS732requestInitFromReply(BYTE_ARRAY_REF_STRUCT(pkt),
			    BYTE_ARRAY_REF_STRUCT(&req));
  req.EEadrs = adrs;
  req.EEwrite = 0;
  req.EEread = 1;

  pkt->EEadrs = -1;
  AVRsendRequest(f,BYTE_ARRAY_REF_STRUCT(&req),BYTE_ARRAY_REF_STRUCT(pkt),
		 &AVREEPROMcallback);
  // printf("EE RD %02X <- [%02X]\n",pkt.pkt.EEdata,adrs);
  req.EEread = 0; // required to return replies to send ADC data
  AVRsendRequest(f,BYTE_ARRAY_REF_STRUCT(&req),BYTE_ARRAY_REF_STRUCT(&pkt2),
		 &AVREEPROMdoneCallback);
  return pkt->EEdata;
}

word AVRcalibrateTime2(FILE *f,word *ts732) {
  int i,min=0xFFFF,max=0,sum=0;
  word fact,hclks = ts732[0x5];

  printf("Calibrating time");
  for (i=0;i<20;i++) {
    hclks = ts732[0x5];
    if (hclks < min) {
      min = hclks;
    }
    if (hclks > max) {
      max = hclks;
    }
    sum += hclks;
    printf("."); fflush(stdout);
    usleep(500000);
  }
  printf("\nMin=%d, Avg=%d, Max=%d\n",min,sum/20,max);

  if (sum / 20 < 0x4000) {
    fact = AVRcalcTimeFactor(100000,0x10000+sum/10);
  } else {
    fact = AVRcalcTimeFactor(100000,sum/10);
  }
  printf("Time factor = %d (0x%X)\n",fact,fact);
  return fact;
}

word AVRcalibrateTime(FILE *f,word *ts732) {
  word fact,hclks = ts732[0x5];

  if (hclks < 0x4000) {
    fact = AVRcalcTimeFactor(100000,0x10000+(dword)hclks*2);
  } else {
    fact = AVRcalcTimeFactor(100000,(dword)hclks*2);
  }
  return fact;
}

// set actual 16-bit value in DAC
void AVRdac1SetValue(word *ts732,word hi) {
  word lo;

  hi >>= 4; // convert 16-bit high time to 12-bit
  lo = 0x1000 - hi;
  ts732[1] = hi;
  ts732[2] = lo;
  return;
}

// voltage: 0 = 0V, 0xFFFF = 10V
void AVRdac1SetCalibratedValue(word *ts732,AVRconfig *conf,word voltage) {
  word DACmax = conf->DAC1max;
  AVRdac1SetValue(ts732,AVRcalcDAChighTime(DACmax,voltage));
}

void AVRdac2SetValue(word *ts732,word hi) {
  word lo;

  hi >>= 4; // convert 16-bit high time to 12-bit
  lo = 0x1000 - hi;
  ts732[3] = hi;
  ts732[4] = lo;
  return;
}

void AVRdac1SetEnable(word *ts732,int en) {
  ts732[0] = (ts732[0] & 0xFE) | (en ? 1 : 0);
}

void AVRdac2SetEnable(word *ts732,int en) {
  ts732[0] = (ts732[0] & 0xFD) | (en ? 2 : 0);
}

void AVRdac2SetCalibratedValue(word *ts732,AVRconfig *conf,word voltage) {
  word DACmax = conf->DAC2max;
  AVRdac2SetValue(ts732,AVRcalcDAChighTime(DACmax,voltage));
}

word AVRcalcDACVoltage(float *v) {
  float f;
  if (*v > 10.0) {
    *v = 10.0;
  }
  if (*v < 0.0) {
    *v = 0;
  }
  f = *v;
  return (word)(f * 65535.0 / 10.0);
}

/*
  Apply 3V (or 4.5V) and read ADC
  Apply 1V (or 0.5V) and read ADC
  calculate offset and scale
  write offset and scale to EEPROM
  save offset and scale in config

  Note: DAC1 loops back to ADC1, and DAC2 looks back to ADC2
  At this point each ADC input is 0-5V.

  n = the schematic numbered ADC to calibrate
 */
void AVRcalibrateADC(FILE *f,word *ts732,int n,AVRconfig *conf) {
  AVRconfig *conf0 = 0;
  AVRreply pkt;
  ByteArrayRef b;
  word ADC2,ADC1,adcTPV,adcOff;
  float volts;
  int i;

  if (n == 1 || n == 3) {
    AVRdac1SetCalibratedValue(ts732,conf,0x7333); // 4.5V
  } else if (n== 2 || n == 4) {
    AVRdac2SetCalibratedValue(ts732,conf,0x7333); // 4.5V
  }
  usleep(200000);
  fileFlushInput(f);
  b = BYTE_ARRAY_REF_STRUCT(&pkt);
  for (i=0;i<20;i++) { // make sure we wait long enough for settling
    b = AVRreplyGet(f,b);
  }
  switch (n) {
  case 1: ADC2 = (pkt.ADC1max + pkt.ADC1min)/2; break;
  case 2: ADC2 = (pkt.ADC2max + pkt.ADC2min)/2; break;
  case 3: ADC2 = (pkt.ADC3max + pkt.ADC3min)/2; break;
  case 4: ADC2 = (pkt.ADC4max + pkt.ADC4min)/2; break;

  }
  printf("ADC1 raw  = %04X:%04X:%04X\n",pkt.ADC1min,pkt.ADC1max,pkt.ADC1);
  printf("ADC2 raw  = %04X:%04X:%04X\n",pkt.ADC2min,pkt.ADC2max,pkt.ADC2);
  printf("ADC3 raw  = %04X:%04X:%04X\n",pkt.ADC3min,pkt.ADC3max,pkt.ADC3);
  printf("ADC4 raw  = %04X:%04X:%04X\n",pkt.ADC4min,pkt.ADC4max,pkt.ADC4);
  printf("--------\n");
  if (n == 1 || n == 3) {
    AVRdac1SetCalibratedValue(ts732,conf,0x0CCD); // 0.5V
  } else if (n == 2 || n == 4) {
    AVRdac2SetCalibratedValue(ts732,conf,0x0CCD); // 0.5V
  } 
  
  usleep(200000);
  fileFlushInput(f);
  b = BYTE_ARRAY_REF_STRUCT(&pkt);
  for (i=0;i<20;i++) { // make sure we wait long enough for settling
    b = AVRreplyGet(f,b);
  }
  switch (n) {
  case 1: ADC1 = (pkt.ADC1max + pkt.ADC1min)/2; break;
  case 2: ADC1 = (pkt.ADC2max + pkt.ADC2min)/2; break;
  case 3: ADC1 = (pkt.ADC3max + pkt.ADC3min)/2; break;
  case 4: ADC1 = (pkt.ADC4max + pkt.ADC4min)/2; break;

  }
  printf("ADC1 raw  = %04X:%04X:%04X\n",pkt.ADC1min,pkt.ADC1max,pkt.ADC1);
  printf("ADC2 raw  = %04X:%04X:%04X\n",pkt.ADC2min,pkt.ADC2max,pkt.ADC2);
  printf("ADC3 raw  = %04X:%04X:%04X\n",pkt.ADC3min,pkt.ADC3max,pkt.ADC3);
  printf("ADC4 raw  = %04X:%04X:%04X\n",pkt.ADC4min,pkt.ADC4max,pkt.ADC4);
  if (ADC2 == ADC1) {
    printf("ADC %d calibration FAILED! (%02X)\n",n,ADC2);
    exit(-1);
  }
  adcTPV = AVRcalcTPV(ADC2,ADC1,9,1);
  adcOff = AVRcalcOff(ADC2,ADC1,9,1);

  if(n == 6) {
    if(conf->ADC4tpv == 0xFFFF || conf->ADC4off == 0xFFFF) {
      printf("ADC4 must be calibrated prior to calibrating ADC6\n");
      exit(2);
    }
      //Based upon resistors, ADC6 = ADC4 * 5.97(total resistance = 36.16k)
      volts = AVRcalcADCvolts(pkt.ADC4,conf->ADC4tpv,conf->ADC4off);	
      adcOff = 0;
      adcTPV = 1 / ((volts * 5.97) / pkt.ADC6);
  }

  switch (n) {
  case 1:
    conf->ADC1tpv = adcTPV;
    conf->ADC1off = adcOff;
    AVRwriteEEPROM(f,(byte)((dword)(&conf0->ADC1tpv)),adcTPV & 0xFF);
    AVRwriteEEPROM(f,((byte)((dword)(&conf0->ADC1tpv)))+1,adcTPV >> 8);
    AVRwriteEEPROM(f,(byte)((dword)(&conf0->ADC1off)),adcOff & 0xFF);
    AVRwriteEEPROM(f,((byte)((dword)(&conf0->ADC1off)))+1,adcOff >> 8);
    break;
  case 2:
    conf->ADC2tpv = adcTPV;
    conf->ADC2off = adcOff;
    AVRwriteEEPROM(f,(byte)((dword)(&conf0->ADC2tpv)),adcTPV & 0xFF);
    AVRwriteEEPROM(f,((byte)((dword)(&conf0->ADC2tpv)))+1,adcTPV >> 8);
    AVRwriteEEPROM(f,(byte)((dword)(&conf0->ADC2off)),adcOff & 0xFF);
    AVRwriteEEPROM(f,((byte)((dword)(&conf0->ADC2off)))+1,adcOff >> 8);
    break;
  case 3:
    conf->ADC3tpv = adcTPV;
    conf->ADC3off = adcOff;
    AVRwriteEEPROM(f,(byte)((dword)(&conf0->ADC3tpv)),adcTPV & 0xFF);
    AVRwriteEEPROM(f,((byte)((dword)(&conf0->ADC3tpv)))+1,adcTPV >> 8);
    AVRwriteEEPROM(f,(byte)((dword)(&conf0->ADC3off)),adcOff & 0xFF);
    AVRwriteEEPROM(f,((byte)((dword)(&conf0->ADC3off)))+1,adcOff >> 8);
    break;
  case 6:
    conf->ADC6tpv = adcTPV;
    conf->ADC6off = adcOff;
    AVRwriteEEPROM(f,(byte)((dword)(&conf0->ADC6tpv)),adcTPV & 0xFF);
    AVRwriteEEPROM(f,((byte)((dword)(&conf0->ADC6tpv)))+1,adcTPV >> 8);
    AVRwriteEEPROM(f,(byte)((dword)(&conf0->ADC6off)),adcOff & 0xFF);
    AVRwriteEEPROM(f,((byte)((dword)(&conf0->ADC6off)))+1,adcOff >> 8);
    break;
  case 4:
    conf->ADC4tpv = adcTPV;
    conf->ADC4off = adcOff;
    AVRwriteEEPROM(f,(byte)((dword)(&conf0->ADC4tpv)),adcTPV & 0xFF);
    AVRwriteEEPROM(f,((byte)((dword)(&conf0->ADC4tpv)))+1,adcTPV >> 8);
    AVRwriteEEPROM(f,(byte)((dword)(&conf0->ADC4off)),adcOff & 0xFF);
    AVRwriteEEPROM(f,((byte)((dword)(&conf0->ADC4off)))+1,adcOff >> 8);
    break;
  }
}

word AVRcalibrateDAC(FILE *f,int n,word *ts732) 
{
	dword upper=0xFFFF, lower = 0, mid;
	AVRreply pkt;
	ByteArrayRef b = BYTE_ARRAY_REF_STRUCT(&pkt);

	int ch = 0;
	AVRconfig *conf = 0;

	if (n != 1 && n != 2) 
		return;

	while (lower < upper) 
	{
		mid = (upper + lower) / 2;
		if (n == 1) 
			AVRdac1SetValue(ts732, mid);
		else 
			AVRdac2SetValue(ts732, mid);
    
		usleep(200000); // settling time
		fileFlushInput(f);
		b = AVRreplyGet(f,b);


		if(n == 1)
		{
			if(pkt.DR8 == 1)
				upper = mid-1;
			else
				lower = mid+1;
		} else
		{

			if(pkt.DR6 == 1)
				upper = mid-1;
			else
				lower = mid+1;
		}
		printf(".");
	}

	printf("\n");
	mid <<= 2;

	// TO DO:restore terminal settings for stdin
	if (n == 1) 
	{
		AVRwriteEEPROM
		  (f,(byte)((dword)(&conf->DAC1max)), mid & 0xFF);
		AVRwriteEEPROM
		  (f,((byte)((dword)(&conf->DAC1max)))+1, mid >> 8);
	} else 
	{
		AVRwriteEEPROM 
		  (f,(byte)((dword)(&conf->DAC2max)), mid & 0xFF);
		AVRwriteEEPROM
		  (f,((byte)((dword)(&conf->DAC2max)))+1, mid >> 8);
	}

	return mid;
}

void AVRreadCalibration(FILE *f,AVRconfig *conf) {
  AVRreply pkt;
  byte *buf = (byte *)conf;
  int i;

  AVRreadEEPROM(f,128,&pkt);
  for (i=0;i<30;i++) {
    buf[128+i] = pkt.EEbytes[i];
  }
}

/*
  EEPROM structure (256 bytes available)
    time calibration
    scaling calibration for each DAC
    scaling and offset calibration for each ADC

  Time calibration:
    timeFactor = ARM_seconds / AVR_seconds

    ARM_seconds = timeFactor * AVR_seconds
    AVR_seconds = ARM_seconds / timeFactor

  DAC calibration:

    highTime = DAC1max * requestedVoltage / maxv
      where: maxv, requestedVoltage
                  0 ->  0V
               4095 -> 10V
	       (~0.244mV increments)
	     DAC1max is scaled the same, but the actual
	     value represents the output required to hit 10.000V

  ADC calibration:

    tpv = ticks per volt
    off = zero offset ticks
    val = raw value read

    ADC1tpv = (adc_ticks_at_V2 - adc_ticks_at_V1) / (V2-V1)
    ADC1off = adc_ticks_at_V1 - ADC1tpv * V1

    ADC1Volts = (ADC1val - ADC1off) / ADC1tpv

  Functions:
    Enable/Disable DAC1
    Enable/Disable DAC2
    LED/Top ON/OFF
    Set Data Direction (5-8)
    Set Data Output Value (5-8)
    Filter for ADC 1,2,3,6,7 (2^n samples psuedo-averaged)
    Shut off for n seconds

  Samples: (each with time)
    ADC (min, max) 1,2,3,6,7
    Data Input Value (5-8)

  (*) not sure we can calibrate these without a specific physical configuration

 */

word *TS732open() {
  int fd = open("/dev/mem",O_RDWR|O_SYNC);
  word *mem;

  if (fd == -1) {
    return 0;
  }
  mem = mmap(0,getpagesize(),PROT_READ|PROT_WRITE, MAP_SHARED,fd,0x72000000);
  if (mem == MAP_FAILED) {
    return 0;
  }
  mem += 0x200; // 400 bytes
  return mem;
}


void TS732setup(FILE *f,void *data) {
  TS732data *t = (TS732data *)data;

  // uncomment next after we see packets ok
  AVRreadCalibration(f,&t->conf);
  if (t->conf.DAC1max == 0xFFFF) {
    printf("Warning: DAC1 is not calibrated\n");
  }
  if (t->conf.DAC2max == 0xFFFF) {
    printf("Warning: DAC2 is not calibrated\n");
  }
  if (t->conf.ADC1tpv == 0xFFFF || t->conf.ADC1off == 0xFFFF) {
    printf("Warning: ADC1 is not calibrated\n");
  }
  if (t->conf.ADC2tpv == 0xFFFF || t->conf.ADC2off == 0xFFFF) {
    printf("Warning: ADC2 is not calibrated\n");
  }
  if (t->conf.ADC3tpv == 0xFFFF || t->conf.ADC3off == 0xFFFF) {
    printf("Warning: ADC3 is not calibrated\n");
  }
  if (t->conf.ADC4tpv == 0xFFFF || t->conf.ADC4off == 0xFFFF) {
    printf("Warning: ADC4 is not calibrated\n");
  }
  if (t->conf.ADC6tpv == 0xFFFF || t->conf.ADC6off == 0xFFFF) {
    printf("Warning: ADC6 is not calibrated\n");
  }
  t->conf.timeFactor = AVRcalibrateTime(f,t->ts732);
}

int TS732requestInitFromReply(ByteArrayRef bRep,ByteArrayRef bReq) {
  AVRrequest *req = (AVRrequest *)bReq.arr;
  AVRreply *reply = (AVRreply *)bRep.arr;

  req->alarm = 0;
  req->outputs = reply->outputs;
  req->DIO = reply->DIO;
  req->ADC2filter = reply->ADC2filter;
  req->ADC3filter = reply->ADC3filter;
  req->ADC4filter = reply->ADC4filter;
  req->ADC6filter = reply->ADC6filter;
  req->ADC1filter = reply->ADC1filter;
  req->version = reply->version;
  return sizeof(AVRrequest);
}

int TS732processOptions(FILE *file,ByteArrayRef bReq,void *data,int *loops,
		       int argc,char **argv) {
  AVRrequest *req = (AVRrequest *)(bReq.arr);
  AVRconfig *conf = &((TS732data *)data)->conf;
  word *ts732 = ((TS732data *)data)->ts732;
  static struct option long_options[] = {
    { "dac1"    ,1,0,0},
    { "dac2"    ,1,0,0},
    { "led1"    ,1,0,0},
    { "led2"    ,1,0,0},
    { "dir5"    ,1,0,0},
    { "dir6"    ,1,0,0},
    { "dir7"    ,1,0,0},
    { "dir8"    ,1,0,0},
    { "data5"   ,1,0,0},
    { "data6"   ,1,0,0},
    { "data7"   ,1,0,0},
    { "data8"   ,1,0,0},
    { "filter1" ,1,0,0},
    { "filter2" ,1,0,0},
    { "filter3" ,1,0,0},
    { "filter6" ,1,0,0},
    { "filter4" ,1,0,0},
    { "sleep"   ,1,0,0},
    { "v1"      ,1,0,0},
    { "v2"      ,1,0,0},
    { "watch"   ,1,0,0},
    { "calibrate-dac", 1,0,0},
    { "eer"     ,1,0,0},
    { "eew"     ,1,0,0},
    { "eeval"   ,1,0,0},
    { "calibrate-adc", 1,0, 0},
    { "help"    ,0,0,0},
    { "calibrate-time", 0, 0, 0},
    { 0,0,0,0 }
  };
  AVRreply pkt;
  int i,c,option_index,doSomething=0;
  word val=0;
  float f;

  *loops = 0;
  opterr = 0; // don't print any messages when we get a bad option
  while (1) {
    c = getopt_long(argc,argv,"",long_options,&option_index);
    if (c == -1) {
      break;
    }
    if (c != 0) {
      continue;
    }
    switch (option_index) {
    case 0: // dac1
      doSomething |= 1;
      req->DAC1en = atoi(optarg);
      AVRdac1SetEnable(ts732,req->DAC1en);
      printf("DAC1: %s\n",req->DAC1en?"enabled":"disabled");
      break;
    case 1: // dac2
      doSomething |= 1;
      req->DAC2en = atoi(optarg);
      AVRdac2SetEnable(ts732,req->DAC2en);
      printf("DAC2: %s\n",req->DAC2en?"enabled":"disabled");
      break;
    case 2: // led1
      doSomething |= 1;
      req->_LEDbot = !atoi(optarg);
      printf("Bottom LED: %s\n",req->_LEDbot?"off":"on");
      break;
    case 3: // led2
      doSomething |= 1;
      req->_LEDtop = !atoi(optarg);
      printf("Top LED: %s\n",req->_LEDtop?"off":"on");
      break;
    case 4: // dir5
      doSomething |= 1;
      req->DDR5 = (optarg[0] == 'I'||optarg[0] == 'i') ? 0 : 1 ;
      printf("Data 5 direction: %s\n",req->DDR5?"output":"input");
      break;
    case 5: // dir6
      doSomething |= 1;
      req->DDR6 = (optarg[0] == 'I'||optarg[0] == 'i') ? 0 : 1 ;
      printf("Data 6 direction: %s\n",req->DDR6?"output":"input");
      break;
    case 6: // dir7
      doSomething |= 1;
      req->DDR7 = (optarg[0] == 'I'||optarg[0] == 'i') ? 0 : 1 ;
      printf("Data 7 direction: %s\n",req->DDR7?"output":"input");
      break;
    case 7: // dir8
      doSomething |= 1;
      req->DDR8 = (optarg[0] == 'I'||optarg[0] == 'i') ? 0 : 1 ;
      printf("Data 8 direction: %s\n",req->DDR8?"output":"input");
      break;
    case 8: // data5
      doSomething |= 1;
      req->DR5 = atoi(optarg);
      printf("Data 5 out: %d\n",req->DR5);
      break;
    case 9: // data6
      doSomething |= 1;
      req->DR6 = atoi(optarg);
      printf("Data 6 out: %d\n",req->DR6);
      break;
    case 10: // data7
      doSomething |= 1;
      req->DR7 = atoi(optarg);
      printf("Data 7 out: %d\n",req->DR7);
      break;
    case 11: // data8
      doSomething |= 1;
      req->DR8 = atoi(optarg);
      printf("Data 8 out: %d\n",req->DR8);
      break;
    case 12: // filter1
      doSomething |= 1;
      req->ADC1filter = atoi(optarg);
      printf("ADC1 filter: 2^%d\n",req->ADC1filter);
      break;
    case 13: // filter2
      doSomething |= 1;
      req->ADC2filter = atoi(optarg);
      printf("ADC2 filter: 2^%d\n",req->ADC2filter);
      break;
    case 14: // filter3
      doSomething |= 1;
      req->ADC3filter = atoi(optarg);
      printf("ADC3 filter: 2^%d\n",req->ADC3filter);
      break;
    case 15: // filter6
      doSomething |= 1;
      req->ADC6filter = atoi(optarg);
      printf("ADC6 filter: 2^%d\n",req->ADC6filter);
      break;
    case 16: // filter7
      doSomething |= 1;
      req->ADC4filter = atoi(optarg);
      printf("ADC4 filter: 2^%d\n",req->ADC4filter);
      break;
    case 17: // sleep
      doSomething |= 1;
      req->alarm = atoi(optarg);
      printf("Sleeping %d seconds\n",req->alarm);
      if (conf->timeFactor != 0xFFFF) {
	req->alarm = AVRcalcAVRseconds(req->alarm,conf->timeFactor);
	printf("AVR seconds to sleep: %d\n",req->alarm);
      }
      break;
    case 18: // v1
      f = atof(optarg);
      if (f < 0.0 || f > 10.0) {
	printf("Error: V1 voltage must be in range 0-10V\n");
      } else {
	AVRdac1SetCalibratedValue(ts732,conf,AVRcalcDACVoltage(&f));
	printf("Setting DAC1 to %fV\n",f);
      }
      break;
    case 19: // v2
      f = atof(optarg);
      if (f < 0.0 || f > 10.0) {
	printf("Error: V2 voltage must be in range 0-10V\n");
      } else {
	AVRdac2SetCalibratedValue(ts732,conf,AVRcalcDACVoltage(&f));
	printf("Setting DAC2 to %fV\n",f);
      }
      break;
    case 20: // watch
      *loops = atoi(optarg);
      break;
    case 21: // calibrate dac
      c = atoi(optarg);
      if (c == 1) {
	AVRdac1SetEnable(ts732,1);
	doSomething |= 3;
      }
      if (c == 2) {
	AVRdac1SetEnable(ts732,1);
	doSomething |= 5;
      }
      break;
    case 22: // eer
      c = atoi(optarg);
      val = AVRreadEEPROM(file,c & 0xFF,&pkt);
      for (i=0;i<30;i++) {
	printf("%02X:%02X\n",(c+i)&0xFF,pkt.EEbytes[i]);
      }
      break;
    case 23: // eew
      c = atoi(optarg);
      printf("Writing %02X to %02X\n",val & 0xFF,c & 0xFF);
      AVRwriteEEPROM(file,c & 0xFF ,val & 0xFF);
      break;
    case 24: // eeval
      val = atoi(optarg);
      break;
    case 25: // calibrate-adc
      c = atoi(optarg);
      if (c == 1) {
	doSomething |= 9;
	req->DAC1en = 1;
	AVRdac1SetEnable(ts732,1);
      }
      if (c == 2) {
	doSomething |= 17;
	req->DAC2en = 1;
	AVRdac2SetEnable(ts732,1);
      }
      if (c == 3) {
	doSomething |= 33;
	req->DAC2en = 1;
	AVRdac2SetEnable(ts732,1);
      }
      if (c == 4) {
	doSomething |= 65;
	req->DAC2en = 1;
	AVRdac2SetEnable(ts732,1);
      }
      if (c == 6) {
        doSomething |= 129;
	req->DAC1en = 0;
	AVRdac1SetEnable(ts732,0);
	req->DAC2en = 0;
	AVRdac2SetEnable(ts732,0);
      }

      break;
    case 26: // help
      printf(
"Usage options:\n"
"  --dac1 [0|1]          disable or enable DAC1\n"
"  --dac2 [0|1]          disable or enable DAC2\n"
"  --led1 [0|1]          turn off or on the bottom LED\n"
"  --led2 [0|1]          turn off or on the top LED\n"
"  --dir5 [I|O]          set pin 5 to an Input or Output\n"
"  --dir6 [I|O]          set pin 6 to an Input or Output\n"
"  --dir7 [I|O]          set pin 7 to an Input or Output\n"
"  --dir8 [I|O]          set pin 8 to an Input or Output\n"
"  --data5 [0|1]         set pin 5 output to low or high\n"
"  --data6 [0|1]         set pin 6 output to low or high\n"
"  --data7 [0|1]         set pin 7 output to low or high\n"
"  --data8 [0|1]         set pin 8 output to low or high\n"
"  --filter1 <n>         set ADC 1 smoothing filter to 2^n (n in 0..15)\n"
"  --filter2 <n>         set ADC 2 smoothing filter to 2^n (n in 0..15)\n"
"  --filter3 <n>         set ADC 3 smoothing filter to 2^n (n in 0..15)\n"
"  --filter6 <n>         set ADC 6 smoothing filter to 2^n (n in 0..15)\n"
"  --filter4 <n>         set ADC 4 smoothing filter to 2^n (n in 0..15)\n"
"  --sleep <n>           sleep n seconds (n in 1..65535)\n"
"  --calibrate-time <n>  calibrate time in n seconds (3000 <= n <= 65535)\n"
"  --calibrate-adc <n>   calibrate ADC n (*)\n"
"  --calibrate-dac <n>   calibrate DAC n (*)\n"
"  --v1 <n>              set DAC1 output to the specific decimal voltage\n"
"  --v2 <n>              set DAC2 output to the specific decimal voltage\n"
"  --watch <n>           skip n-1 messages from AVR, then print current state\n"
"  --eer <adrs>          display 30 bytes of AVR EEPROM contents, starting at n\n"
"  --eeval <val>         set the value to write to the EEPROM (testing only)\n"
"  --eew <adrs>          write value to given AVR EEPROM adrs (testing only)\n"
"  --calibrate-time      perform an more accurate time calibration this run\n"
"  --help                get this help message\n"
      );
      exit(0);
    case 27: // calibrate-time
      AVRcalibrateTime2(file,ts732);
      break;
    }
  }
  if ((doSomething & 1) == 0 && *loops == 0) {
    *loops = 1;
  }
  return doSomething;
}

void TS732replyPacket_Print(ByteArrayRef bRep,void *data) {
  AVRreply *pkt = (AVRreply *)bRep.arr;
  AVRconfig *conf = &((TS732data *)data)->conf;
  float f1,f2;

  // TO DO: calibrate ADC readings to decimal voltages
  printf("----------------\n");
  printf("time      = %d.%03d\n",pkt->millitime/1000,pkt->millitime%1000);
  AVRadcVoltsPrint(1,pkt->ADC1,pkt->ADC1min,pkt->ADC1max,conf->ADC1tpv,conf->ADC1off);
  AVRadcVoltsPrint(2,pkt->ADC2,pkt->ADC2min,pkt->ADC2max,conf->ADC2tpv,conf->ADC2off);
  AVRadcVoltsPrint(3,pkt->ADC3,pkt->ADC3min,pkt->ADC3max,conf->ADC3tpv,conf->ADC3off);
  AVRadcVoltsPrint(4,pkt->ADC4,pkt->ADC4min,pkt->ADC4max,conf->ADC4tpv,conf->ADC4off);
  AVRadcVoltsPrint(6,pkt->ADC6,pkt->ADC6min,pkt->ADC6max,conf->ADC6tpv,conf->ADC6off);
  printf("DAC1 EN   = %d\n",pkt->DAC1en);
  printf("DAC2 EN   = %d\n",pkt->DAC2en);
  //printf("reserved0 = %d\n",pkt->rsvd0);
  printf("LED bot   = %s\n",pkt->_LEDbot ? "OFF" : "ON");
  printf("LED top   = %s\n",pkt->_LEDtop ? "OFF" : "ON");
  printf("EE write  = %d\n",pkt->EEwrite);
  printf("EE read   = %d\n",pkt->EEread);
  //printf("reserved1 = %d\n",pkt->rsvd1);
  //printf("DR5 Intr  = %d\n",pkt->DR5IntGen);
  //printf("reserved3 = %d\n",pkt->rsvd3);
  //printf("reserved4 = %d\n",pkt->rsvd4);
  printf("EE ready  = %d\n",pkt->EEready);
  //printf("reserved2 = %d\n",pkt->rsvd2);
  printf("DDR5      = %d\n",pkt->DDR5);
  printf("DDR6      = %d\n",pkt->DDR6);
  printf("DDR7      = %d\n",pkt->DDR7);
  printf("DDR8      = %d\n",pkt->DDR8);

  //IN5 doesn't have an inverter on the input pin although
  //IN6,IN7,and IN8 do. we invert in software to simplify the
  //user interface.
  if(pkt->DDR5==0)
    printf("DR5       = %d\n",(pkt->DR5) ^ 1);
  else
    printf("DR5       = %d\n",pkt->DR5);

  printf("DR6       = %d\n",pkt->DR6);
  printf("DR7       = %d\n",pkt->DR7);
  printf("DR8       = %d\n",pkt->DR8);
  printf("EE adrs   = %d\n",pkt->EEadrs);
  printf("EE data   = %d\n",pkt->EEdata);
  printf("ADC2 filt = %d\n",pkt->ADC2filter);
  printf("ADC3 filt = %d\n",pkt->ADC3filter);
  printf("ADC4 filt = %d\n",pkt->ADC4filter);
  printf("ADC6 filt = %d\n",pkt->ADC6filter);
  printf("ADC1 filt = %d\n",pkt->ADC1filter);
  //printf("version   = %d\n",pkt->version);
}


void TS732taskDo(FILE *f,int todo,ByteArrayRef bRep,void *data) {
  TS732data *t = (TS732data *)data;

  if (todo & 2) {
    printf("Calibrating DAC %d\n",1);
    AVRcalibrateDAC(f,1,t->ts732);
  }
  if (todo & 4) {
    printf("Calibrating DAC %d\n",2);
    AVRcalibrateDAC(f,2,t->ts732);
  }
  if (todo & 8) {
      printf("Calibrating ADC %d\n",1);
      AVRcalibrateADC(f,t->ts732,1,&t->conf);
  }
  if (todo & 16) {
      printf("Calibrating ADC %d\n",2);
      AVRcalibrateADC(f,t->ts732,2,&t->conf);
  }
  if (todo & 32) {
      printf("Calibrating ADC %d\n",3);
      AVRcalibrateADC(f,t->ts732,3,&t->conf);
  }
  if (todo & 64) {
      printf("Calibrating ADC %d\n",4);
      AVRcalibrateADC(f,t->ts732,4,&t->conf);
  }
  if (todo & 128) {
      printf("Calibrating ADC %d\n", 6);
      AVRcalibrateADC(f,t->ts732,6,&t->conf);
  }
}

int TS732callback(ByteArrayRef bReq,ByteArrayRef bRep) {
  AVRrequest *req = (AVRrequest *)bReq.arr;
  AVRreply *rep = (AVRreply *)bRep.arr;

  return (rep->DAC1en != req->DAC1en
	  || rep->DAC2en != req->DAC2en
	  || rep->_LEDbot != req->_LEDbot
	  || rep->_LEDtop != req->_LEDtop
	  || rep->DDR5 != req->DDR5
	  || rep->DDR6 != req->DDR6
	  || rep->DDR7 != req->DDR7
	  || rep->DDR8 != req->DDR8
	  /*
	  || req->DDR5 ? (rep->DR5 != req->DR5) : 1
	  || req->DDR6 ? (rep->DR6 != req->DR6) : 1
	  || req->DDR7 ? (rep->DR7 != req->DR7) : 1
	  || req->DDR8 ? (rep->DR8 != req->DR8) : 1
	  */
	  || rep->ADC2filter != req->ADC2filter
	  || rep->ADC3filter != req->ADC3filter
	  || rep->ADC4filter != req->ADC4filter
	  || rep->ADC6filter != req->ADC6filter
	  || rep->ADC1filter != req->ADC1filter) ? 1 : 0;
}

int main(int argc,char *argv[]) {
  TS732data data;

  data.ts732 = TS732open();
  if (!data.ts732) {
    perror("TS732open:");
    return 1;
  }

  AVRrun("/dev/ttyTS0",&data,argc,argv,&TS732setup,
	 &TS732requestInitFromReply,&TS732processOptions,
	 &TS732replyPacket_Print,&TS732taskDo,&TS732callback);
  return 0;
}

