#include <stdlib.h>
#include "filter.h"

// copyright (c)2006 Technologic Systems
// Author: Michael Schmidt

int filterDataSmoothAverageFloat_Push(filter *f,void *din);
void *filterDataSmoothAverageFloat_Pull(filter *f,void *dout);
int filterDataSchmittTriggerFloat_Push(filter *f,void *din);
void *filterDataSchmittTriggerFloat_Pull(filter *f,void *dout);
int filterDataRateFloat_Push(filter *f,void *din);
void *filterDataRateFloat_Pull(filter *f,void *dout);

// ***************
// FILTER: GENERIC
// ***************
inline int filterPush(filter *f,void *din) {
  return (f && f->push) ? f->push(f,din) : 0;
}

inline void *filterPull(filter *f,void *dout) {
  return (f && f->pull) ? 
    f->pull(f,dout) : 0;
}

inline filter *filterDo(filter *filt,filter *prev) {
  if (prev) filt->tmp = prev->tmp;
  return (prev && filterPush(filt,filterPull(prev,filt->tmp))) ? filt : 0;
}

inline filter *filterDo0(filter *filt,void *din) {
  filt->tmp = din;
  return filterPush(filt,din) ? filt : 0;
}

// ************************
// FILTER: SMOOTHED AVERAGE
// ************************
void filterDataSmoothAverageFloat_Init(filter *f,int maxSamples) {
  filterDataSmoothAverageFloat *A;

  f->push = &filterDataSmoothAverageFloat_Push;
  f->pull = &filterDataSmoothAverageFloat_Pull;
  f->data = malloc(sizeof(filterDataSmoothAverageFloat));
  A = f->data;
  A->numSamples = 0;
  A->maxSamples = maxSamples;
  A->data = malloc(sizeof(float)*maxSamples);
}

int filterDataSmoothAverageFloat_Push(filter *f,void *din) {
  filterDataSmoothAverageFloat *A = f->data;
  float value = ((float *)din)[0];
  int i;

  if (A->numSamples < A->maxSamples) {
    A->data[A->numSamples++] = value;
  } else {
    for (i=0;i<A->maxSamples-1;i++) {
      A->data[i] = A->data[i+1];
    }
    A->data[i] = value;
  }
  return A->numSamples >= 2; //  A->maxSamples / 2);
}

void *filterDataSmoothAverageFloat_Pull(filter *f,void *dout) {
  filterDataSmoothAverageFloat *A;
  int i;
  float *value = dout;
  float min, max;
  int  minI, maxI, n;
  float Sum;
  int j;

  if (!f || !f->data) {
    return 0;
  }
  A = f->data;
  if (A->numSamples == 0) {
    return 0;
  }
  if (A->numSamples == 1) {
    value[0] = A->data[0];
  } else if (A->numSamples == 2) {
    value[0] = (A->data[0] + A->data[1]) / 2;
  } else {
    min = max = A->data[0];
    minI = maxI = 0;
    for (j=1;j<A->numSamples;j++) {
      if (A->data[j] < min) {
	minI = j;
	min = A->data[j];
      }
      if (A->data[j] > max) {
	maxI = j;
	max = A->data[j];
      }
    }
    n = 0; Sum = 0;
    for (j=0;j<A->numSamples;j++) {
      if (j != minI && j != maxI) {
	Sum += A->data[j];
	n++;
      }
    }
    //assert(n > 0.5);
    value[0] = Sum / n;
  }
  return dout;
}

// ***********************
// FILTER: SCHMITT TRIGGER
// ***********************

void filterDataSchmittTriggerFloat_Init(filter *f,float hi,float lo) {
  filterDataSchmittTriggerFloat *S;

  f->push = &filterDataSchmittTriggerFloat_Push;
  f->pull = &filterDataSchmittTriggerFloat_Pull;
  f->data = malloc(sizeof(filterDataSchmittTriggerFloat));  
  S = f->data;
  S->highWater = hi;
  S->lowWater = lo;
  S->passingData = 0;
}

/*
  value > highWater starts passing the data until value < lowWater
 */
int filterDataSchmittTriggerFloat_Push(filter *f,void *din) {
  filterDataSchmittTriggerFloat *S = f->data;
  float value = ((float *)din)[0];

  if (S->passingData) {
    S->passingData = (value >= S->lowWater);
  } else {
    S->passingData = (value > S->highWater);
  }
  if (S->passingData) {
    S->lastValue = value;
  }
  return S->passingData;
}

void *filterDataSchmittTriggerFloat_Pull(filter *f,void *dout) {
  filterDataSchmittTriggerFloat *S = f->data;
  float *value = dout;

  if (S->passingData) {
    value[0] = S->lastValue;
    return dout;
  } else {
    return 0;
  }
}

// ************
// FILTER: RATE
// ************

/*
  The Rate filter measures the rate of change.
  It does this by keeping the past N samples, and then looking at the
  difference between the first sample and the last sample.

  The more samples we use, the lower the rate of change we can detect,
  at the expense of being able to detect transients.
 */
void filterDataRateFloat_Init(filter *f,int maxSamples) {
  filterDataSmoothAverageFloat *A;

  f->push = &filterDataRateFloat_Push;
  f->pull = &filterDataRateFloat_Pull;
  f->data = malloc(sizeof(filterDataSmoothAverageFloat));
  A = f->data;
  A->numSamples = 0;
  A->maxSamples = maxSamples;
  A->data = malloc(sizeof(float)*maxSamples);
}

int filterDataRateFloat_Push(filter *f,void *din) {
  filterDataSmoothAverageFloat *A = f->data;

  filterDataSmoothAverageFloat_Push(f,din);
  return A->numSamples > 1;
}

void *filterDataRateFloat_Pull(filter *f,void *dout) {
  filterDataSmoothAverageFloat *A;
  float *value = dout;

  if (!f || !f->data) {
    return 0;
  }
  A = f->data;
  if (A->numSamples < 2) {
    return 0;
  }
  value[0] = (A->data[A->numSamples-1] - A->data[0]);
  return dout;
}
