#include <sys/ioctl.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include "mtd.h"

/*
  TO DO:
  Add bounds checking on all front-level accesses, and abort if we go
  past the bounds of the device.

  MTDread
  MTDwrite
 */
int getMTDInfo(FILE *ff,mtd_info_t *inf) {
  if (ioctl(fileno(ff), MEMGETINFO, inf) == -1) {
    perror("ioctl:MEMGETINFO");
    return 0;
  }
  return 1;
}

inline int MTDoffset(MTDFILE *f,int block,int sector) {
  return f->info.erasesize * block +  f->info.oobblock * sector;
}

inline int MTDblock(MTDFILE *f,long offset) {
  return offset / f->info.erasesize;
}

inline int MTDsector(MTDFILE *f,long offset) {
  return (offset / f->info.oobblock) % (f->info.erasesize/f->info.oobblock);
}

inline int MTDbyte(MTDFILE *f,long offset) {
  return (offset % f->info.oobblock);
}

inline int MTDseek(MTDFILE *f,int block,int sector,int byte) {
  return (fseek(f->f,byte+MTDoffset(f,block,sector),SEEK_SET) == 0);
}

int MTDreadOOB(MTDFILE *f,int block,int sector,char *oob) {
  struct mtd_oob_buf oob_buf;
  int ret;

  if (!MTDseek(f,block,sector,0)) {
    return 0;
  }

  oob_buf.length = f->info.oobsize;
  oob_buf.ptr = oob;
  oob_buf.start = MTDoffset(f,block,sector);

  if (ioctl(fileno(f->f), MEMREADOOB, &oob_buf) == -1 ||
      ferror(f->f)) {
    clearerr(f->f);
    return 0;
  }
  return 1;
}

int MTDeraseBlock(MTDFILE *f,int block) {
  int ret;
  erase_info_t erase;
  char oob[f->info.oobsize];

  erase.start = block * f->info.erasesize;
  erase.length = f->info.erasesize;

  /*
  if (!MTDreadOOB(f,block,0,oob) || oob[5] != 0xFF) {
    return 0; // don't try to erase if block is marked bad
  }
  */
  if (ioctl(fileno(f->f), MEMERASE, &erase) == -1
      || ferror(f->f)) {
    clearerr(f->f);
    return 0;
  }
  return 1;
}

void MTDinvalidateBlock(MTDFILE *f,int block) {
  struct mtd_oob_buf oob_buf;
  unsigned char oob[f->info.oobsize];

  MTDeraseBlock(f,block); // try to erase it first
  oob_buf.length = f->info.oobsize;
  oob_buf.ptr = oob;
  oob_buf.start = f->info.erasesize * block;

  memset(oob,0,f->info.oobsize);
  ioctl(fileno(f->f), MEMWRITEOOB, &oob_buf);
}

int MTDreadBlock(MTDFILE *f,char *buf,int len) {
  int rc;

  if (ftell(f->f) >= f->info.size) {
    printf("MTDreadBlock past end of device, aborting\n");
    exit(3);
  }
  if ((rc=fread(buf,len,1,f->f)) != 1
      || ferror(f->f)) {
    clearerr(f->f);
    return 0;
  } else {
    return 1;
  }
}

int MTDwriteBlock(MTDFILE *f,char *buf,int len) {
  if (ftell(f->f)+len >= f->info.size) {
    printf("MTDreadBlock past end of device, aborting\n");
    exit(3);
  }
  if (fwrite(buf,len,1,f->f) != 1
      || ferror(f->f)) {
    clearerr(f->f);
    return 0;
  } else {
    return 1;
  }
}

int MTDverifyBlock(MTDFILE *f,char *buf,int len) {
  char buf2[len];

  if (fseek(f->f,-len,SEEK_CUR) != 0) {
    return 0;
  }
  if (fread(buf2,len,1,f->f) != 1) {
    return 0;
  }
  return (memcmp(buf,buf2,len) == 0);
}

int MTDread(MTDFILE *f,char *buf,int len) {
  int offset,block,sector,byte,len0;
  static int maxerr = 6;

  offset = ftell(f->f);
  block = MTDblock(f,offset);
  sector = MTDsector(f,offset);
  byte = MTDbyte(f,offset);
  if (sector != 0 || byte != 0) { // offset must be block aligned
    return 0;
  }
  while (len > 0) {
    len0 = len > f->info.erasesize ? f->info.erasesize : len;
    if (!MTDreadBlock(f,buf,len0)) {
      printf("read: skipping bad block %d\n",block);
      if (maxerr-- <= 0) {
	printf("Too many errors -- aborting\n");
	exit(3);
      }
      block++;
      if (!MTDseek(f,block,0,0)) {
	return 0;
      }
      continue;
    }
    len -= len0;
    buf += len0;
  }
  return 1;
}

int MTDwrite(MTDFILE *f,char *buf,int len) {
  int offset,block,sector,byte,len0;
  static int maxerr = 6;

  offset = ftell(f->f);
  block = MTDblock(f,offset);
  sector = MTDsector(f,offset);
  byte = MTDbyte(f,offset);
  if (sector != 0 || byte != 0) { // offset must be block aligned
    return 0;
  }
  while (len > 0) {
    //len0 = len > f->info.erasesize ? f->info.erasesize : len;
    len0 = f->info.erasesize;
    //printf("Writing %ld for %d\n",ftell(f->f),len);
    if (!MTDeraseBlock(f,block) || 
	!MTDwriteBlock(f,buf,len0) || 
	!MTDverifyBlock(f,buf,len0)) {
      printf("write: skipping bad block %d\n",block);
      if (maxerr-- <= 0) {
	printf("Too many errors -- aborting\n");
	exit(3);
      }
      MTDinvalidateBlock(f,block);
      block++;
      if (!MTDseek(f,block,0,0)) {
	return 0;
      }
      continue;
    } 
    len -= len0;
    buf += len0;
    block++;
  }
  return 1;
}

MTDFILE *MTDopen(char *name) {
  MTDFILE *f = malloc(sizeof(MTDFILE));

  f->f = fopen(name,"r+");
  if (f->f == NULL) {
    free(f);
    return 0;
  }
  if (!getMTDInfo(f->f,&f->info)) {
    fclose(f->f);
    free(f);
    return 0;
  }
  setvbuf(f->f,0,_IONBF,0);
  return f;
}

void MTDclose(MTDFILE *f) {
  if (!f) {
    return;
  }
  fclose(f->f);
  free(f);
}
